/*
 * Copyright (c) 2021 Chris Camacho (codifies -  https://bedroomcoders.co.uk/)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "raylib.h"
#include "raymath.h"

#define RLIGHTS_IMPLEMENTATION
#include "rlights.h"

// the initial window size, also used returning from full screen
#define windowWidth 1280
#define windowHeight 720

#define max( a, b ) ( ( a ) > ( b ) ? ( a ) : ( b ) )
#define min( a, b ) ( ( a ) < ( b ) ? ( a ) : ( b ) )

int main(void)
{
    //------------------------------------------------------------------
    // Initialization
    //------------------------------------------------------------------
    SetTraceLogLevel( LOG_ALL );
    SetConfigFlags( FLAG_WINDOW_RESIZABLE | FLAG_VSYNC_HINT | FLAG_MSAA_4X_HINT );
    InitWindow( windowWidth, windowHeight, "Raylib - template" );

    // the screen size we render to, that is then scaled
    int virtualWidth = 0, virtualHeight = 0;
    {
        int m = GetCurrentMonitor();
        virtualWidth = GetMonitorWidth( m );
        virtualHeight = GetMonitorHeight( m );
    }

    // Define the camera to look into our 3d world
    Camera camera = { 0 };
    camera.position = (Vector3){ 0.0f, 1.0f, 4.0f };
    camera.target = (Vector3){ 0.0f, 0.0f, 0.0f };
    camera.up = (Vector3){ 0.0f, 1.0f, 0.0f };
    camera.fovy = 45.0f;
    camera.type = CAMERA_PERSPECTIVE;

    // RTT target for the virtual screen
    RenderTexture2D target = LoadRenderTexture( virtualWidth, virtualHeight );
    SetTextureFilter( target.texture, FILTER_BILINEAR );
    
    // lighting shader
    Shader shader = LoadShader( "data/simpleLight.vs", "data/simpleLight.fs" );
    shader.locs[SHADER_LOC_MATRIX_MODEL] = GetShaderLocation( shader, "matModel" );
    shader.locs[SHADER_LOC_VECTOR_VIEW] = GetShaderLocation( shader, "viewPos" );

    // ambient light level
    int amb = GetShaderLocation( shader, "ambient" );
    SetShaderValue( shader, amb, (float[4]){0.2,0.2,0.2,1.0}, SHADER_UNIFORM_VEC4 );

    // make a light (max 4 but we're only using 1)
    Light light = CreateLight( LIGHT_POINT, (Vector3){ 2,4,1 }, Vector3Zero(), WHITE, shader );

    // set the shader for each material in the model
    Model model = LoadModel( "data/axis.gltf" );
    for ( int i = 0; i < model.materialCount; i++ ) {
        model.materials[i].shader = shader;
    }
    
    // model rotation
    Vector3 ang = { 0 };

    // setting screen size AND toggling fullscreen needs to be done separately
    int fsToggle = 0;
    
    // delta time
    float dT = 0;
    
    // for the sake of the bit not covered by the render texture
    ClearBackground( BLACK );

    //------------------------------------------------------------------
    // Main game loop
    //------------------------------------------------------------------

    while ( !WindowShouldClose() )    // Detect window close button or ESC key
    {
        //--------------------------------------------------------------
        // Update
        //--------------------------------------------------------------
        dT = GetFrameTime();

        // adjust mouse coordinates for scale and offset
        float screenScale = min( (float)GetScreenWidth() / virtualWidth, (float)GetScreenHeight() / virtualHeight );
        
        SetMouseScale( 1.0 / screenScale, 1.0 / screenScale );
        SetMouseOffset( -( GetScreenWidth() - ( virtualWidth * screenScale ) ) * 0.5,
                        -( GetScreenHeight() - ( virtualHeight * screenScale ) ) * 0.5 );

        // code by doris the dog, can't see why it won't compile for now
        // nhm
        
        // handle fullscreen transition
        if ( fsToggle ) {
            fsToggle--;
            if ( !fsToggle ) {
                ToggleFullscreen();
                // need to set window size here or it upsets toggling
                if ( !IsWindowFullscreen() ) {
                    SetWindowSize( windowWidth, windowHeight );
                    SetWindowPosition( (GetScreenWidth() - windowWidth)/2,
                        (GetScreenHeight() - windowHeight)/2 );
                }
            }
        }
        
        if ( IsKeyDown( KEY_F1 ) ) {
            fsToggle = 2;   // at least two or toggling doesn't work
            if ( !IsWindowFullscreen() ) {
                SetWindowSize( virtualWidth, virtualHeight );
            }
        }
    
        // update the model
        ang.x += 1.0f * dT;
        ang.y += .5 * dT;
        ang.z -= 2.5 * dT;
        
        model.transform = MatrixRotateZYX( ang );

        // update the camera and lighting shader
        // not actually needed unless you move the camera
        // but here so it doesn't get forgotten
        UpdateCamera( &camera );

        // position the light slightly above the camera
        light.position = camera.position;
        light.position.y += 0.1f;

        // update the light shader with the camera view position
        SetShaderValue( shader, shader.locs[SHADER_LOC_VECTOR_VIEW], &camera.position.x, SHADER_UNIFORM_VEC3 );
        UpdateLightValues( shader, light );

        //--------------------------------------------------------------
        // Draw
        //--------------------------------------------------------------
        
        BeginDrawing();

        // clears the window outside the drawing area if needed
        if (IsWindowResized() ) {
            ClearBackground( BLACK );
        }

        // render to a texture so we can scale to the window        
        BeginTextureMode( target );
        ClearBackground( (Color){64,128,255,255} );
        
        // 3D rendering
        BeginMode3D( camera );

        DrawModel( model, (Vector3){0,0,0}, 1, WHITE );
        DrawGrid( 10, 1.0f );

        EndMode3D();

        // 2D overlay
        DrawText( FormatText("screen scale %.4f", screenScale ), 20, 80, 40, WHITE );
        DrawText( "F1 to toggle fullscreen", 20, 120, 40, WHITE );
        
        // haven't hidden the mouse cursor to verify its right!
        Vector2 mp = GetMousePosition();
        DrawCircle( mp.x,mp.y,12,RED );
        
        EndTextureMode();

        // draw the virtual screen on the actual one
        DrawTexturePro( target.texture, (Rectangle) { 0.0f, 0.0f,
            target.texture.width,  -target.texture.height },
            (Rectangle) { ( GetScreenWidth() - virtualWidth * screenScale ) * 0.5,
                ( GetScreenHeight() - virtualHeight * screenScale ) * 0.5,
                virtualWidth*screenScale, virtualHeight*screenScale },
                    (Vector2) { 0, 0 }, 0.0f, WHITE );


        EndDrawing();
    }

    //------------------------------------------------------------------
    // De-Initialization
    //------------------------------------------------------------------

    UnloadModel( model );
    UnloadShader( shader );

    // Close window and OpenGL context
    CloseWindow();
    
    return 0;
}



